
/*
?- dynamic elems/2.
?- dynamic brianbug/0.
*/

elems(4,[a,b,c,d]).   

go :- 
	initialize(Tables, Comm),
	interind(Tables, Comm).



initialize(Tables, Example) :-
	writelist([nl,'Interleaving inductive inference  * 1 *',nl,nl]),
	writelist(['Enter first example: ']),
	read(E),
	const_to_list(E, Example),
	retract(elems(_,_)),
	size(Example,N),
	assert(elems(N,Example)),
	writelist([nl,'Creating tables...',nl]),
	create_perm_tables(Example, Tables),
	!.



interind(Tables, Comm) :- 
	writelist(['Constructing expression...',nl]),
	cleancomm(Comm, CleanComm),
	create_expr(Tables, Expr, CleanComm),
	brianbug(Tables),
	present((Expr,CleanComm), Answer),
	iterate(Tables, Comm, Answer).
interind(Tables, Comm) :-
	writelist(['*** Expressions exhausted.',nl]),
	present(new_example, Answer),
	iterate(Tables, Comm, Answer).

bug :- assert(brianbug).
nobug :- retract(brianbug).




cleancomm([], []).
cleancomm([no(_)|R], S) :- !, cleancomm(R, S).
cleancomm([A|R], [A|S]) :- cleancomm(R, S).



set_nocomm([], []).
set_nocomm([no(A)|R], [no(A)|S]) :- 
	!,
	set_nocomm(R, S).
set_nocomm([A|R], [no(A)|S]) :- 
	set_nocomm(R, S).




iterate(_, _, [e,n,d]) :- !.
iterate(_, _, [m,o,r,e]) :- 
	!,
	fail.
iterate(Tables, Comm, [n,o,c,o,m,m]) :- 
	!,
	set_nocomm(Comm, Comm2),
	interind(Tables, Comm2).
iterate(Tables, Comm, Example) :-
	writelist(['Processing example...',nl]),
	process_example(Example, Tables, NewTables),
	process_comm(Example, Comm, NewComm),
	interind(NewTables, NewComm).



present(new_example, Answer) :-
	repeat,
	writef('end or counterexample: '),
	read(A),
	const_to_list(A, Answer),
	check_input(Answer),
	!.
present((Expr,Comm), Answer) :-
	repeat,
	writelist([nl,'Expression --> ', Expr]),
	writelist([nl,'      Comm  =  ', Comm,nl,nl]),
	writef('end, more, nocomm, or counterexample: '),
	read(A),
	const_to_list(A, Answer),
	check_input(Answer),
	!.



brianbug(Tables) :-
	brianbug,
	prtgraph(Tables),
	!.
brianbug(_).

check_input([e,n,d]) :- !.
check_input([m,o,r,e]) :- !.
check_input([n,o,c,o,m,m]) :- !.
check_input(A) :-
	elems(_,E),
	\+ subset(A,E),
	!,
	writelist([nl,'Error: illegal elements in input.',nl,nl]),
	fail.
check_input(A) :-
	elems(_,E),
	\+ subset(E,A),
	!,
	writelist([nl,'Error: missing elements in input.',nl,nl]),
	fail.
check_input(A) :-
	elems(N, _),
	size(A, M),
	N > M,
	!,
	writelist([nl,'Error: example too small.',nl,nl]),
	fail.
check_input(A) :-
	elems(N, _),
	size(A, M),
	N < M,
	!,
	writelist([nl,'Error: example too large.',nl,nl]),
	fail.
check_input(_).






create_expr(Tables, Expr, Comm) :-
	elems(_, U),
	create_expr_terms(Tables, U, Comm, Elist),
	convert_expr(Elist, Expr).









create_expr_terms(Tables, U, [], Elist) :-
	!,
	create_nocomm_expr(Tables, U, Elist).

create_expr_terms(Tables, U, Comm, Elist) :-  
	get_comm_seq(Tables, U, Comm, Comm1, Tleft, Uleft, Tright, Uright),
	create_nocomm_expr(Tleft, Uleft, Elistleft),
	remove_list(Comm1, Comm, Comm2),
	create_expr_terms(Tright, Uright, Comm2, Elistright),
	merge_terms(Elistleft, Comm1, Elistright, Elist),
	!.







get_comm_seq(Tables, U, [A|R], [A|Seq], Tleft, Uleft, Tright, Uright) :-
	get_1st_comm_elem(Tables, A, Tleft, Uleft, Tables2),
	get_rest_comm_seq(Tables2, R, Seq, Tables3),
	appendD(Uleft, [A|Seq], Tmp),
	filter(Tmp, Tables3, Tright),
	remove_list(Tmp, U, Uright).






get_1st_comm_elem(Tables, A, Tleft, Uleft, Tables2) :-
	append(X, [table(_,Arcs)|Y], Tables),
	member([A,_], Arcs),
	!,
	filter([A], X, X2),
	relativize(X2, Tleft),
	get_elems(X, UX),
	remove_all(A, UX, Uleft),
	filter(Uleft, Y, Tables2).
get_1st_comm_elem(Tables, A, Tables2, Uleft, []) :- 
	filter([A], Tables, Tables2),
	get_elems(Tables2, Uleft).





get_rest_comm_seq([], Comm, Comm, []) :- !.
get_rest_comm_seq(Tables, [], [], Tables) :- !.
get_rest_comm_seq([table(_,Arcs)|Rest], [A|R], [A|S], Tables2) :- 
	member([A,_], Arcs),
	!,
	get_rest_comm_seq(Rest, R, S, Tables2).
get_rest_comm_seq(Tables, _, [], Tables) :- !.






merge_terms([], Comm, [], [Comm]) :- !.
merge_terms([], Comm, Right, Elist) :-
	!,
	prepend_all(Comm, Right, Elist).
merge_terms(Left, Comm, [], Elist) :-
	!,
	append_all(Left, Comm, Elist).
merge_terms(Left, Comm, Right, Elist) :-
	sort_size(Left, Left2),
	sort_size(Right, Right2),
	size(Left2, Lsize),
	size(Right2, Rsize),
	(Lsize = Rsize ->
		append_all(Left2, Comm, Left3),
		match(Left3, Right2, Elist)
		;
	( Lsize > Rsize ->
		append_all(Left2, Comm, Left3),
		match(Left3, Right2, Elist)
		;
		prepend_all(Comm, Right2, Right3),
		match(Left2, Right3, Elist))).




match([], B, B) :-!.
match(A, [], A) :-!.
match([A|R], [B|S], [C|T]) :-
	appendD(A, B, C),
	match(R, S, T).



sort_size(Termlist, Sorted) :-
	add_sizes(Termlist, Sizelist),
	revsort(Sizelist, SizeSorted),
	remove_sizes(SizeSorted, Sorted),
	!.

add_sizes([], []).
add_sizes([A|R],[(N,A)|S]) :- 
	size(A, N),
	add_sizes(R, S).

remove_sizes([], []).
remove_sizes([(_,A)|R], [A|S]) :-
	remove_sizes(R, S).






create_nocomm_expr(Tables, U, Elist) :-
	construct_nocomm_list(Tables, U, [], Elist).

construct_nocomm_list([], [], [], []) :- !.
construct_nocomm_list([], Elems, List, FinalList) :- 
	!,
	degenerate_list(Elems, List, FinalList).
construct_nocomm_list(_, [], List, List) :- !.

construct_nocomm_list(Tables, Elems, List, FinalList) :-
	gen_largest_term(_, Tables, Elems, TermList),
	remove_list(TermList, Elems, Elems2),
	size(Elems2, N),
	(N > 1 -> 
		filter(TermList, Tables, Tables2),
		append(_, [table(M,T)|Rest], Tables2), 
		M =< N,                       
		TempTables = [table(M,T)|Rest]	     
		; 
		TempTables = []),
	construct_nocomm_list(TempTables, Elems2, [TermList|List], FinalList).
construct_nocomm_list(_, Elems, List, FinalList) :-  
	!,
	degenerate_list(Elems, List, FinalList).

degenerate_list([], E, E).
degenerate_list([A|B], E, [[A]|E2]) :-
	degenerate_list(B, E, E2).



gen_largest_term(A, [], Elems, [A]) :- 
	member(A, Elems).
gen_largest_term(A, [table(_,Arcs)|LowerTables], Elems, [A|Tlist]) :-
	member([A,B], Arcs),
	append(X,[A|Y],Elems),
	appendD(X,Y, Elems2),
	gen_largest_term(B, LowerTables, Elems2, Tlist).

remove_elems([], E, E).
remove_elems([A|As], E, NewE) :-
	remove_all(A, E, E2),
	!,
	remove_elems(As, E2, NewE).

get_elems([], []).
get_elems(Tables, Elems) :-
	append(_,[table(_,Arcs)], Tables),
	flatten(Arcs, Arcs2),
	remove_dups(Arcs2, Elems),
	!.

flatten([],[]).
flatten([[A,B]|R],[A,B|S]) :- flatten(R, S).

bumpen([], []).
bumpen([A|R], [[A]|S]) :- bumpen(R, S).

extract_comm_list(A, [table(_,[[A,B]])|Y], [A|RestComm], RestTable) :- 
	extract_comm_list2(B, Y, RestComm, RestTable).

extract_comm_list2(_, [], [], []).
extract_comm_list2(A, [table(_,[[A,B]])|Y], [A|RestComm], RestTable) :- 
	extract_comm_list2(B, Y, RestComm, RestTable).
extract_comm_list2(_, [_|RestTable], [], RestTable).




filter(S, T, NT) :-
	reverse(T, RT),
	filter2(S, RT, RT2),
	reverse(RT2, NT),
	!.

filter2([], T, T).
filter2([A|R], T, NT) :-
	filter3(A, T, T3),
	filter2(R, T3, NT).

filter3(_, [], []) :- !.
filter3(_, [table(_,[])|_], []) :- !.
filter3(A, [table(N,Arcs)|R], NT) :-
	member([A,_], Arcs),
	!,
	remove_all([A,_], Arcs, Arcs2),
	filter3(A, [table(N,Arcs2)|R], NT).
filter3(A, [table(N,Arcs)|R], NT) :-
	member([B,A], Arcs),
	!,
	remove([B,A], Arcs, Arcs2),
	(\+ member([B,_], Arcs2) ->
		filter3(B, R, R2)
		;
		R = R2),
	filter3(A, [table(N,Arcs2)|R2], NT).
filter3(A, [T|R], [T|NT]) :-
	filter3(A, R, NT).


relativize([], []).
relativize(Tables, RTables) :-
	appendD(_, [table(N,_)], Tables),
	D is N - 2,
	set_tab(Tables, D, RTables).

set_tab([], _, []).
set_tab([table(N,X)|R], D, [table(M,X)|R2]) :-
	M is N - D,
	set_tab(R, D, R2).


convert_expr([], '').
convert_expr([Termlist], Term) :-
	!,
        list_to_const(Term, Termlist).
convert_expr([Termlist|Rest], Term2#Term) :-
	list_to_const(Term, Termlist),
	convert_expr(Rest, Term2).



gen_perms(0, _, P, P).
gen_perms(N, Example, Perms, FinalPerms) :-
	N > 0,
	perms(N, Example, Perms3),
	append(Perms, Perms3, Perms4),
	N2 is N - 1,
	gen_perms(N2, Example, Perms4, FinalPerms).


perms(N, E, Perms) :-
	setof(P,permute(N, E, P), Perms).

permute(N, E, _) :- 
	(N < 1 ; size(E,M), N > M),
	!,
	fail.
permute(1,E,[A]) :-
	member(A,E).
permute(N,[E|Es],[E|B]) :-
	N > 1,
	N2 is N - 1,
	permute(N2, Es, B).
permute(N, [_|Es], A) :-
	N > 1,
	permute(N, Es, A).

create_perm_tables(Example, Tables) :-
	elems(N, _),
	gen_perm_tables(Example, N, [], Tables2),
	revsort(Tables2, Tables),
	!.

gen_perm_tables(_, 1, T, T).
gen_perm_tables(Ex, N, Tables, FinalTables) :-
	N > 1,
	elems(Tot,_),
	K is Tot-N+2,
	append(Es, _, Ex),
	size(Es, K),
	perms(2, Es, Perms),
	N2 is N - 1,
	gen_perm_tables(Ex, N2, [table(N, Perms)|Tables], FinalTables).
	


process_example(Example, Tables, NewTables) :-
	perms(2, Example, Pairs),
	conflicts(Pairs, Tables, Tables2),
	clean_tables(Tables2, Tables3),
	revsort(Tables3, NewTables),
	!.

conflicts([], T, T).
conflicts([[A,B]|R], Tables, NewTables) :-
	table_entry(2, [B,A], Tables),
	!,
	del_all_entries([B,A], Tables, [], Tables2),
	conflicts(R, Tables2, NewTables).
conflicts([_|R], Tables, NewTables) :-
	conflicts(R, Tables, NewTables).

table_entry(N, E, Tables) :-
	memberD(table(N,TableN), Tables),
	member(E, TableN),
	!.

del_all_entries(_,[],T,T) :- !.
del_all_entries(E,[table(N,Table)|Rest],Processed, NewTables) :-
	\+ \+ member(E, Table),
	!,
	remove(E, Table, Table2),
	del_all_entries(E,Rest, [table(N,Table2)|Processed],Tables2),
	process_null_row(E,table(N,Table2),Tables2, NewTables).
del_all_entries(E, [T|Rest], Processed, NewTables) :-
	del_all_entries(E,Rest,[T|Processed], NewTables).

del_column(_, _, [], T, T) :- !.
del_column(N, B, [table(N,Table)|Rest], Processed, NewTables) :-
	member([A,B], Table),
	!,
	remove([A,B], Table, Table2),
	del_column(N, B, [table(N,Table2)|Rest], Processed, Tables3),
	process_null_row([A,B], table(N,Table2), Tables3, NewTables).
del_column(N, E, [T|Rest], Processed, NewTables) :-
	del_column(N, E, Rest, [T|Processed], NewTables).

process_null_row([A,_], table(N,Table), Tables, NewTables) :-
	(\+(member([A,_], Table)); Table=[]),
	!,
	N2 is N + 1,
	del_column(N2, A, Tables, [], NewTables).
process_null_row(_, _, Tables, Tables).

clean_tables([],[]).
clean_tables([table(_,[])|Rest], Tables) :-
	!,
	clean_tables(Rest, Tables).
clean_tables([T|Rest],[T|Tables]) :-
	clean_tables(Rest,Tables).








process_comm(E, C, C2) :-
	process_comm2(E, C, [], [], C2).

process_comm2([], [], _, _, []).
process_comm2([A|R], [no(B)|Y], S, T, [no(B)|W]) :-
	!,
	append(S, [A], S2),
	append(T, [B], T2),
	process_comm2(R, Y, S2, T2, W).
process_comm2([A|R], [A|Y], S, T, [A|W]) :-
	subset(S, T),
	!,
	append(S, [A], S2),
	append(T, [A], T2),
	process_comm2(R, Y, S2, T2, W).
process_comm2([A|R], [B|Y], S, T, [no(B)|W]) :-
	!,
	append(S, [A], S2),
	append(T, [B], T2),
	process_comm2(R, Y, S2, T2, W).
	




merge_terms(A, [], A) :- !.
merge_terms([], B, B).
merge_terms(A, B, C) :-
	merge_terms2(A,B,C).

merge_terms2([], _, []).
merge_terms2([A|R], B, T) :-
	merge_terms3(A, B, C),
	merge_terms2(R, B, S),
	append(C, S, T).

merge_terms3(_, [], []).
merge_terms3(A, [B|R], [C|S]) :-
	appendD(A, B, C),
	merge_terms3(A, R, S).



prtgraph([]) :- nl.
prtgraph([table(N,Arcs)|R]) :-
	writelist([nl, 'T',N, ':']),
	prtgraph2(Arcs),
	prtgraph(R).

prtgraph2([]) :- !.
prtgraph2([[A,B]|R]) :-
	nl,
	tab(4),
	writelist([A,' -> ',B]),
	prtgraph2(R).


